"""Calculate the cost of a completion using the Bedrock API."""


def calculate_cost(
    input_tokens: int | float | None, output_tokens: int | float | None, model: str
) -> float | None:
    """Calculate the cost of a completion using the Bedrock API."""
    # NOTE: We are currently investigating a dynamic approach to determine costs
    # for Bedrock models. This is needed due to the large number of models
    # available through Bedrock, making it impractical to hardcode all pricing
    # information in this initial implementation.
    return None



from anthropic import AsyncAnthropic
from mirascope.core import BaseMessageParam, anthropic


@anthropic.call("claude-3-5-sonnet-20240620", client=AsyncAnthropic())
async def recommend_book_async(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



from mirascope.core import BaseMessageParam, cohere


@cohere.call("command-r-plus", call_params={"max_tokens": 512})
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


response = recommend_book("fantasy")
print(response.content)



import asyncio

from mirascope.core import BaseMessageParam, groq


@groq.call("llama-3.1-70b-versatile")
async def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


async def main():
    response = await recommend_book("fantasy")
    print(response.content)


asyncio.run(main())



from vertexai.generative_models import FunctionDeclaration, GenerativeModel, Tool

model = GenerativeModel("gemini-1.5-flash")


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    response = model.generate_content(
        f"Who wrote {book}?",
        tools=[
            Tool(
                function_declarations=[
                    FunctionDeclaration(
                        **{
                            "name": "get_book_author",
                            "description": "Returns the author of the book with the given title.",
                            "parameters": {
                                "properties": {"title": {"type": "string"}},
                                "required": ["title"],
                                "type": "object",
                            },
                        }
                    )
                ]
            )
        ],
    )
    if tool_calls := [
        part.function_call
        for candidate in response.candidates  # pyright: ignore [reportAttributeAccessIssue]
        for part in candidate.content.parts
        if part.function_call.args
    ]:
        if tool_calls[0].name == "get_book_author":
            return get_book_author(**dict(tool_calls[0].args.items()))  # pyright: ignore [reportArgumentType]
    return response.text  # pyright: ignore [reportAttributeAccessIssue]


author = identify_author("The Name of the Wind")
print(author)



from mirascope.core import anthropic


@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import Messages, openai


@openai.call("gpt-4o-mini")
def summarize(text: str) -> Messages.Type:
    return Messages.User(f"Summarize this text: {text}")


@openai.call("gpt-4o-mini")
def summarize_and_translate(text: str, language: str) -> Messages.Type:
    summary = summarize(text)
    return Messages.User(f"Translate this text to {language}: {summary.content}")


response = summarize_and_translate("Long English text here...", "french")
print(response.content)



from google.generativeai import GenerationConfig
from mirascope.core import gemini, prompt_template


@gemini.call(
    "gemini-1.5-flash",
    call_params={"generation_config": GenerationConfig(max_output_tokens=512)},
)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



"""The Mirascope Azure Module."""

from typing import TypeAlias

from azure.ai.inference.models import ChatRequestMessage

from ..base import BaseMessageParam
from ._call import azure_call
from ._call import azure_call as call
from .call_params import AzureCallParams
from .call_response import AzureCallResponse
from .call_response_chunk import AzureCallResponseChunk
from .dynamic_config import AzureDynamicConfig
from .stream import AzureStream
from .tool import AzureTool, AzureToolConfig

AzureMessageParam: TypeAlias = ChatRequestMessage | BaseMessageParam

__all__ = [
    "call",
    "AzureDynamicConfig",
    "AzureCallParams",
    "AzureCallResponse",
    "AzureCallResponseChunk",
    "AzureMessageParam",
    "AzureStream",
    "AzureTool",
    "AzureToolConfig",
    "azure_call",
]



"""Tests the `openai.stream` module."""

import pytest
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import (
    Choice as ChunkChoice,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.completion_usage import CompletionUsage

from mirascope.core.openai.call_response import OpenAICallResponse
from mirascope.core.openai.call_response_chunk import OpenAICallResponseChunk
from mirascope.core.openai.stream import OpenAIStream
from mirascope.core.openai.tool import OpenAITool


def test_openai_stream() -> None:
    """Tests the `OpenAIStream` class."""
    assert OpenAIStream._provider == "openai"

    class FormatBook(OpenAITool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    tool_call_delta = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunks = [
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(content="content", tool_calls=None), index=0
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[tool_call_delta],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
            usage=usage,
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
            if tool_calls := call_response_chunk.chunk.choices[0].delta.tool_calls:
                assert tool_calls[0].function
                tool_call = ChatCompletionMessageToolCall(
                    id="id",
                    function=Function(**tool_calls[0].function.model_dump()),
                    type="function",
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = OpenAIStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=OpenAICallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost == 1.25e-05

    format_book = FormatBook.from_tool_call(
        ChatCompletionMessageToolCall(
            id="id",
            function=Function(
                arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
                name="FormatBook",
            ),
            type="function",
        )
    )
    assert stream.message_param == {
        "role": "assistant",
        "content": "content",
        "tool_calls": [format_book.tool_call.model_dump()],
    }


def test_construct_call_response() -> None:
    """Tests the `OpenAIStream.construct_call_response` method."""

    class FormatBook(OpenAITool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    tool_call = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunks = [
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(content="content", tool_calls=None), index=0
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
            usage=usage,
        ),
    ]

    tool_call = None

    def generator():
        nonlocal tool_call
        for chunk in chunks:
            call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
            if tool_calls := call_response_chunk.chunk.choices[0].delta.tool_calls:
                assert tool_calls[0].function
                tool_call = ChatCompletionMessageToolCall(
                    id="id",
                    function=Function(**tool_calls[0].function.model_dump()),
                    type="function",
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = OpenAIStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=OpenAICallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
    )

    for _ in stream:
        pass

    tool_call = ChatCompletionMessageToolCall(
        id="id",
        function=Function(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    completion = ChatCompletion(
        id="id",
        choices=[
            Choice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content="content", role="assistant", tool_calls=[tool_call]
                ),
            )
        ],
        created=0,
        model="gpt-4o",
        object="chat.completion",
        usage=usage,
    )
    call_response = OpenAICallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.response == call_response.response


def test_construct_call_response_no_usage() -> None:
    """Tests the `OpenAIStream.construct_call_response` method with no usage."""
    chunks = [
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(content="content", tool_calls=None), index=0
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = OpenAIStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=OpenAICallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
    )
    for _ in stream:
        pass
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.usage is None


def test_openai_stream_audio() -> None:
    """Tests the `OpenAIStream` class with audio output."""
    chunks = [
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="content",
                        audio={"id": "audio-id-123"},  # pyright: ignore [reportCallIssue]
                    ),
                    index=0,
                )
            ],
            created=0,
            model="gpt-4o",
            object="chat.completion.chunk",
        ),
    ]

    def generator():
        for chunk in chunks:
            yield OpenAICallResponseChunk(chunk=chunk), None

    stream = OpenAIStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=OpenAICallResponse,
        model="gpt-4o",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
    )

    # Stream should initially have no audio_id
    assert stream.audio_id is None

    # Consume stream
    for _ in stream:
        pass

    # Check audio_id was set
    assert stream.audio_id == "audio-id-123"

    # Verify audio_id is included in message param
    assert stream.message_param == {
        "role": "assistant",
        "content": "content",
        "audio": {"id": "audio-id-123"},
    }



"""Utility for determining if a prompt template has been decorated."""

from collections.abc import Awaitable, Callable
from typing import ParamSpec

from typing_extensions import TypeIs

from ..dynamic_config import BaseDynamicConfig
from ..messages import Messages

_P = ParamSpec("_P")


def is_prompt_template(
    fn: Callable[..., BaseDynamicConfig | Messages.Type]
    | Callable[..., Awaitable[BaseDynamicConfig | Messages.Type]],
) -> TypeIs[
    Callable[..., BaseDynamicConfig]
    | Callable[
        ...,
        Awaitable[BaseDynamicConfig],
    ]
]:
    return hasattr(fn, "__mirascope_prompt_template__")



from typing import Literal

from mirascope.core import BaseMessageParam, anthropic, openai, prompt_template
from mirascope.core.base import BaseCallResponse
from pydantic import BaseModel


class Librarian(BaseModel):
    provider: Literal["openai", "anthropic"]
    model: Literal["gpt-4o-mini", "claude-3-5-sonnet-20240620"]
    history: list[BaseMessageParam] = []

    @prompt_template()
    def _prompt(self, query: str) -> list[BaseMessageParam]:
        return [
            BaseMessageParam(role="system", content="You are a librarian"),
            *self.history,
            BaseMessageParam(role="user", content=(query)),
        ]

    def _call(self, query: str) -> BaseCallResponse:
        if self.provider == "openai":
            call = openai.call(self.model)(self._prompt)
        elif self.provider == "anthropic":
            call = anthropic.call(self.model)(self._prompt)
        else:
            raise ValueError(f"Unsupported provider: {self.provider}")
        return call(query)

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            response = self._call(query)
            print(response.content)
            self.history += [
                BaseMessageParam(role="user", content=query),
                response.message_param,
            ]


Librarian(provider="openai", model="gpt-4o-mini").run()



from pydub import AudioSegment

from mirascope.core import prompt_template, Messages


@prompt_template()
def identify_book_prompt(audio_wave: AudioSegment) -> Messages.Type:
    return Messages.User(
        ["Here's an audio book snippet:", audio_wave, "What book is this?"]
    )


with open("....", "rb") as audio:
    print(identify_book_prompt(AudioSegment.from_mp3(audio)))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="Here's an audio book snippet:"),
#             AudioPart(type='audio', media_type='audio/wav', audio=b'...'),
#             TextPart(type="text", text="What book is this?"),
#         ],
#     )
# ]



import asyncio
from typing import Any, Callable

from mirascope.beta.openai import Context, Realtime, async_input

from mirascope.beta.openai import OpenAIRealtimeTool

app = Realtime(
    "gpt-4o-realtime-preview-2024-10-01",
    modalities=["text"],
)


def format_book(title: str, author: str) -> str:
    return f"{title} by {author}"


@app.sender(wait_for_text_response=True)
async def send_message(context: Context) -> tuple[str, list[Callable]]:
    genre = await async_input("Enter a genre: ")
    return f"Recommend a {genre} book. please use the tool `format_book`.", [
        format_book
    ]


@app.receiver("text")
async def receive_text(response: str, context: dict[str, Any]) -> None:
    print(f"AI(text): {response}", flush=True)


@app.function_call(format_book)
async def recommend_book(tool: OpenAIRealtimeTool, context: Context) -> str:
    result = tool.call()
    print(result)
    return result


asyncio.run(app.run())



"""Tests the `bedrock._utils.get_json_output` module."""

from typing import Any, cast

import pytest

from mirascope.core.bedrock._types import StreamOutputChunk
from mirascope.core.bedrock._utils._get_json_output import get_json_output
from mirascope.core.bedrock.call_params import BedrockCallParams
from mirascope.core.bedrock.call_response import BedrockCallResponse
from mirascope.core.bedrock.call_response_chunk import BedrockCallResponseChunk


def test_get_json_output_call_response() -> None:
    """Tests the `get_json_output` function with a call response."""
    message: dict[str, Any] = {
        "role": "assistant",
        "content": [
            {
                "text": '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            }
        ],
    }
    call_response = BedrockCallResponse(
        metadata={},  # pyright: ignore [reportArgumentType]
        response={  # pyright: ignore [reportArgumentType]
            "output": {"message": message},
        },
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        call_params=BedrockCallParams(max_tokens=1000),  # pyright: ignore [reportCallIssue]
        call_kwargs={},  # pyright: ignore [reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
        messages=[],
    )
    assert (
        get_json_output(call_response, json_mode=True)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    message["content"] = [
        {
            "type": "toolUse",
            "toolUse": {
                "input": {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
            },
        }
    ]
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    message["content"] = [{"type": "toolUse", "toolUse": {}}]
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)

    # Test case for empty message content
    message["content"] = []
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)

    # Test case for missing message
    call_response.response = {"output": {}}  # pyright: ignore [reportAttributeAccessIssue]
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk() -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    chunk = {
        "contentBlockDelta": {
            "delta": {
                "text": "json_output",
            }
        }
    }
    call_response_chunk = BedrockCallResponseChunk(chunk=chunk)  # pyright: ignore [reportArgumentType]
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"

    chunk["contentBlockDelta"]["delta"] = {  # pyright: ignore [reportArgumentType]
        "toolUse": {
            "input": '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
        }
    }
    assert (
        get_json_output(call_response_chunk, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    chunk["contentBlockDelta"]["delta"] = {"text": ""}
    assert get_json_output(call_response_chunk, json_mode=False) == ""

    # Test case for missing contentBlockDelta
    chunk = cast(StreamOutputChunk, {})
    call_response_chunk = BedrockCallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=False) == ""

    # Test case for missing delta in contentBlockDelta
    chunk = cast(StreamOutputChunk, {"contentBlockDelta": {}})
    call_response_chunk = BedrockCallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=False) == ""



from mirascope.core import prompt_template
from pydantic import BaseModel


class Book(BaseModel):
    title: str
    author: str


@prompt_template()
def recommend_book_prompt(book: Book) -> str:
    return f"I read {book.title} by {book.author}. What should I read next?"


book = Book(title="The Name of the Wind", author="Patrick Rothfuss")
print(recommend_book_prompt(book))
# Output: [BaseMessageParam(role='user', content='I read The Name of the Wind by Patrick Rothfuss. What should I read next?')]



from mirascope.core import Messages, groq


def parse_recommendation(response: groq.GroqCallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@groq.call("llama-3.1-70b-versatile", output_parser=parse_recommendation)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book. Output only Title by Author")


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



from mirascope.core import mistral, prompt_template
from mistralai.async_client import MistralAsyncClient


@mistral.call("mistral-large-latest", client=MistralAsyncClient())
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



from mirascope.core import BaseDynamicConfig, BaseMessageParam, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> BaseDynamicConfig:
    uppercase_genre = genre.upper()
    messages = [
        BaseMessageParam(role="user", content=f"Recommend a {uppercase_genre} book")
    ]
    return {
        "messages": messages,
        "computed_fields": {"uppercase_genre": uppercase_genre},
    }


print(recommend_book_prompt("fantasy"))
print(recommend_book_prompt("fantasy"))
# Output: {
#     "messages": [BaseMessageParam(role="user", content="Recommend a FANTASY book")],
#     "computed_fields": {"uppercase_genre": "FANTASY"},
# }



from mirascope.core import BaseDynamicConfig, openai, prompt_template


@openai.call("gpt-4o-mini")
@prompt_template("Summarize this text: {text}")
def summarize(text: str): ...


@openai.call("gpt-4o-mini")
@prompt_template("Translate this text to {language}: {summary}")
def summarize_and_translate(text: str, language: str) -> BaseDynamicConfig:
    return {"computed_fields": {"summary": summarize(text)}}


response = summarize_and_translate("Long English text here...", "french")
print(response.content)



from mirascope.core import BaseDynamicConfig, Messages, openai, prompt_template
from pydantic import BaseModel


class Librarian(BaseModel):
    history: list[openai.OpenAIMessageParam] = []

    def _ask_for_help(self, question: str) -> str:
        """Asks for help from an expert."""
        print("[Assistant Needs Help]")
        print(f"[QUESTION]: {question}")
        answer = input("[ANSWER]: ")
        print("[End Help]")
        return answer

    @openai.call("gpt-4o-mini")
    @prompt_template(
        """
        SYSTEM: You are a librarian
        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _call(self, query: str) -> BaseDynamicConfig:
        return {"tools": [self._ask_for_help]}

    def _step(self, query: str) -> str:
        if query:
            self.history.append(Messages.User(query))
        response = self._call(query)
        self.history.append(response.message_param)
        tools_and_outputs = []
        if tools := response.tools:
            for tool in tools:
                print(f"[Calling Tool '{tool._name()}' with args {tool.args}]")
                tools_and_outputs.append((tool, tool.call()))
            self.history += response.tool_message_params(tools_and_outputs)
            return self._step("")
        else:
            return response.content

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            step_output = self._step(query)
            print(step_output)


Librarian().run()



import asyncio
from typing import Any

from mirascope.beta.openai import Context, Realtime, async_input

app = Realtime(
    "gpt-4o-realtime-preview-2024-10-01",
    modalities=["text"],
)


@app.receiver("text")
async def receive_text(response: str, context: dict[str, Any]) -> None:
    print(f"AI(text): {response}", flush=True)


@app.sender(wait_for_text_response=True)
async def send_message(context: Context) -> str:
    message = await async_input("Enter your message: ")
    return message


asyncio.run(app.run())



"""Tests the `_utils.get_metadata` module."""

from mirascope.core.base._utils._get_metadata import get_metadata
from mirascope.core.base.metadata import Metadata
from mirascope.core.base.prompt import BasePrompt, metadata


def test_get_metadata() -> None:
    """Tests the `get_metadata` function."""

    data: Metadata = {"tags": {"version:0001"}}

    class Prompt(BasePrompt):
        """prompt"""

    assert get_metadata(Prompt, None) == {}
    assert get_metadata(metadata(data)(Prompt), None) == data
    assert get_metadata(Prompt, {"metadata": {"tags": {"version:0002"}}}) == {
        "tags": {"version:0002"}
    }

    def fn() -> None:
        """prompt"""

    assert get_metadata(fn, None) == {}
    assert get_metadata(metadata(data)(fn), None) == data
    assert get_metadata(fn, {"metadata": {"tags": {"version:0002"}}}) == {
        "tags": {"version:0002"}
    }



"""Tests the `anthropic._utils.setup_call` module."""

import inspect
from unittest.mock import MagicMock, patch

import pytest
from anthropic import Anthropic

from mirascope.core.anthropic._utils._setup_call import setup_call
from mirascope.core.anthropic.tool import AnthropicTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch(
    "mirascope.core.anthropic._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.anthropic._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {"max_tokens": 1000}
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="claude-3-5-sonnet-20240620",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={"max_tokens": 1000},
        extract=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert (
        "model" in call_kwargs and call_kwargs["model"] == "claude-3-5-sonnet-20240620"
    )
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, AnthropicTool, {"max_tokens": 1000}
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    assert inspect.signature(create) == inspect.signature(Anthropic().messages.create)


@patch("mirascope.core.anthropic._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_system_message(
    mock_utils: MagicMock, mock_base_setup_call: MagicMock
) -> None:
    mock_base_setup_call.return_value[1] = [
        {"role": "system", "content": [{"type": "text", "text": "test"}]}
    ]
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {"max_tokens": 1000}
    _, _, _, _, call_kwargs = setup_call(
        model="claude-3-5-sonnet-20240620",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={"max_tokens": 1000},
        extract=False,
    )
    assert "system" in call_kwargs
    assert call_kwargs["system"] == [{"type": "text", "text": "test"}]


@patch(
    "mirascope.core.anthropic._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.anthropic._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock()
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"type": "text", "text": "test"}]}
    ]
    mock_base_setup_call.return_value[3] = {"max_tokens": 1000, "tools": MagicMock()}
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="claude-3-5-sonnet-20240620",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={"max_tokens": 1000},
        extract=False,
    )
    assert messages[-1]["content"][-1] == {  # type: ignore
        "type": "text",
        "text": mock_utils.json_mode_content.return_value,
    }
    assert "tools" not in call_kwargs

    mock_utils.json_mode_content.return_value = "\n\njson"
    mock_base_setup_call.return_value[1] = [{"role": "user", "content": "test"}]
    _, _, messages, _, call_kwargs = setup_call(
        model="claude-3-5-sonnet-20240620",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={"max_tokens": 1000},
        extract=False,
    )
    assert messages[-1]["content"] == "test\n\njson"  # type: ignore


@patch(
    "mirascope.core.anthropic._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.anthropic._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""
    mock_base_setup_call.return_value[2] = [
        MagicMock(spec=AnthropicTool, __name__="tool")
    ]
    mock_base_setup_call.return_value[3] = {"max_tokens": 1000}
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, tool_types, call_kwargs = setup_call(
        model="claude-3-5-sonnet-20240620",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={"max_tokens": 1000},
        extract=True,
    )
    assert isinstance(tool_types, list)
    assert "tool_choice" in call_kwargs and call_kwargs["tool_choice"] == {
        "type": "tool",
        "name": tool_types[0]._name(),
    }



import logfire
from mirascope.core import gemini
from mirascope.integrations.logfire import with_logfire
from pydantic import BaseModel

logfire.configure()


class Book(BaseModel):
    title: str
    author: str


@with_logfire()
@gemini.call("gemini-1.5-flash", response_model=Book)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import Messages, bedrock
import boto3


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0", client=boto3.client("bedrock-runtime")
)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")
